#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Included data files for standalone mode.

This keeps track of data files for standalone mode. Do not should them for
DLLs or extension modules, these need to be seen by Nuitka as entry points
for dependency analysis.
"""

import fnmatch
import os

from nuitka import ModuleRegistry
from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.containers.OrderedSets import OrderedSet
from nuitka.importing.Importing import locateModule
from nuitka.options.BuildPackageCommon import getProjectExpectedDataFiles
from nuitka.options.Options import (
    getOutputPath,
    getShallIncludeDataDirs,
    getShallIncludeDataFiles,
    getShallIncludeExternallyDataFilePatterns,
    getShallIncludePackageData,
    getShallIncludeRawDirs,
    getShallNotIncludeDataFilePatterns,
    isAcceleratedMode,
    isStandaloneMode,
    shallMakeModule,
)
from nuitka.OutputDirectories import getStandaloneDirectoryPath
from nuitka.plugins.Hooks import considerDataFiles, onDataFileTags
from nuitka.PythonFlavors import getSystemPrefixPath
from nuitka.PythonVersions import (
    getSitePackageCandidateNames,
    python_version_str,
)
from nuitka.Tracing import general, inclusion_logger, options_logger
from nuitka.utils.FileOperations import (
    addFileExecutablePermission,
    areSamePaths,
    containsPathElements,
    copyFileWithPermissions,
    getFileContents,
    getFileList,
    getFilenameExtension,
    getFileSize,
    getNormalizedPath,
    isFilenameBelowPath,
    isLegalPath,
    isRelativePath,
    makePath,
    openTextFile,
    relpath,
    resolveShellPatternToFilenames,
)
from nuitka.utils.Importing import getExtensionModuleSuffixes
from nuitka.utils.Utils import getArchitecture, isAIX, isMacOS

data_file_tags = []


def addDataFileTags(pattern):
    """Add a data file tag pattern.

    Args:
        pattern: The pattern and tags in "pattern:tags" format.
    """
    assert ":" in pattern, pattern
    data_file_tags.append(pattern)


def getDataFileTags(dest_path):
    """Get the data file tags for a specific destination path.

    Args:
        dest_path: The destination path to check.

    Returns:
        OrderedSet of tags applicable to the path.
    """
    result = OrderedSet()

    for value in data_file_tags:
        pattern, tags = value.rsplit(":", 1)

        if fnmatch.fnmatch(dest_path, pattern):
            result.update(tags.split(","))

    return result


def hasDataFileTag(tag):
    """Check if a specific tag exists in any registered patterns.

    Args:
        tag: The tag to search for.

    Returns:
        bool: True if the tag is found, False otherwise.
    """
    for value in data_file_tags:
        _pattern, tags = value.rsplit(":", 1)

        if tag in tags:
            return True


def decodeDataFileTags(tags):
    """Convert tags argument to an OrderedSet.

    Notes:
        In many places, strings are accepted for tags, convert to OrderedSet
        for internal use.

    Args:
        tags: string (comma separated) or iterable of tags.

    Returns:
        OrderedSet of tags.
    """

    if type(tags) is str:
        tags = tags.split(",") if tags else ()

    return OrderedSet(tags)


class IncludedDataFile(object):
    """Details of an included data file.

    Notes:
        This class holds all information about a data file to be included in
        the standalone distribution.

    Args:
        kind: The kind of data file ("data_file" or "data_blob").
        source_path: The source path of the file (if kind is "data_file").
        dest_path: The destination path in the distribution.
        reason: The reason for inclusion.
        data: The binary data (if kind is "data_blob").
        tags: Tags associated with this file.
        tracer: The tracer to use for logging.
    """

    __slots__ = ("kind", "source_path", "dest_path", "reason", "data", "tags", "tracer")

    def __init__(self, kind, source_path, dest_path, reason, data, tags, tracer):
        tags_set = getDataFileTags(dest_path)

        current = dest_path
        while True:
            current = os.path.dirname(current)
            if not current:
                break

            tags_set.update(getDataFileTags(current))

        tags_set.update(decodeDataFileTags(tags))

        # Copy, unless specified otherwise.
        if not any(tag.startswith("embed-") for tag in tags_set):
            tags_set.add("copy")

        self.kind = kind
        self.source_path = source_path
        self.dest_path = getNormalizedPath(dest_path)

        is_legal, illegal_reason = isLegalPath(self.dest_path)
        if not is_legal:
            general.sysexit(
                "Error, cannot add data file with '%s' path, as '%s'"
                % (dest_path, illegal_reason)
            )

        self.reason = reason
        self.data = data
        self.tags = tags_set
        self.tracer = tracer

    def __repr__(self):
        return "<%s %s source '%s' dest '%s' reason '%s' tags '%s'>" % (
            self.__class__.__name__,
            self.kind,
            self.source_path,
            self.dest_path,
            self.reason,
            ",".join(self.tags),
        )

    def needsCopy(self):
        """Check if the file needs to be copied.

        Returns:
            bool: True if the file should be copied, False otherwise.
        """
        return "copy" in self.tags

    def getFileContents(self):
        """Get the contents of the data file.

        Returns:
            bytes: The content of the file.
        """
        if self.kind == "data_file":
            return getFileContents(filename=self.source_path, mode="rb")
        elif self.kind == "data_blob":
            return self.data
        else:
            assert False

    def getFileSize(self):
        """Get the size of the data file.

        Returns:
            int: The size of the file in bytes.
        """
        if self.kind == "data_file":
            return getFileSize(self.source_path)
        elif self.kind == "data_blob":
            return len(self.data)
        else:
            assert False


def makeIncludedEmptyDirectory(dest_path, reason, tracer, tags):
    """Create an included data file representing an empty directory.

    Args:
        dest_path: The destination path for the directory.
        reason: The reason for inclusion.
        tracer: The tracer to use for logging.
        tags: Tags associated with this directory.

    Returns:
        IncludedDataFile: The created data file object (using a marker file).
    """
    dest_path = os.path.join(dest_path, ".keep_dir.txt")

    return makeIncludedGeneratedDataFile(
        data="", dest_path=dest_path, reason=reason, tracer=tracer, tags=tags
    )


def makeIncludedDataFile(source_path, dest_path, reason, tracer, tags):
    """Create an included data file.

    Args:
        source_path: The source path of the file.
        dest_path: The destination path in the distribution.
        reason: The reason for inclusion.
        tracer: The tracer to use for logging.
        tags: Tags associated with this file.

    Returns:
        IncludedDataFile: The created data file object.
    """
    tags = decodeDataFileTags(tags)

    # Refuse directories, these must be kept distinct.
    if os.path.isdir(source_path):
        tracer.sysexit(
            "Error, cannot include directory '%s' as a data file. Data directories have their own options."
            % source_path
        )

    # In accelerated mode, data files can be everywhere, but they cannot
    # change place.
    if isAcceleratedMode():
        if "copy" in tags:
            if "package_data" not in tags and not areSamePaths(source_path, dest_path):
                tracer.sysexit(
                    "Error, cannot change paths for data files in accelerated mode from '%s' to '%s'."
                    % (source_path, dest_path)
                )
    else:
        if not isRelativePath(dest_path):
            tracer.sysexit(
                "Error, cannot use dest path '%s' outside of distribution." % dest_path
            )

    return IncludedDataFile(
        kind="data_file",
        source_path=source_path,
        dest_path=dest_path,
        data=None,
        reason=reason,
        tracer=tracer,
        tags=tags,
    )


# By default ignore all things that are code.
default_ignored_suffixes = (
    ".py",
    ".pyw",
    ".pyc",
    ".pyo",
    ".pyi",
    ".so",
    ".pyd",
    ".pyx",
    ".dll",
    ".dylib",
    ".exe",
    ".bin",
)

default_ignored_suffixes += getExtensionModuleSuffixes()

default_ignored_dirs = ("__pycache__",) + getSitePackageCandidateNames()

default_ignored_filenames = ("py.typed",)
if not isMacOS():
    default_ignored_filenames += (".DS_Store",)


def makeIncludedDataDirectory(
    source_path,
    dest_path,
    reason,
    tracer,
    tags,
    ignore_dirs=(),
    ignore_filenames=(),
    ignore_suffixes=(),
    only_suffixes=(),
    normalize=True,
    raw=False,
):
    """Create included data files from a directory.

    Args:
        source_path: The source directory path.
        dest_path: The destination path in the distribution.
        reason: The reason for inclusion.
        tracer: The tracer to use for logging.
        tags: Tags associated with these files.
        ignore_dirs: Directories to ignore.
        ignore_filenames: Filenames to ignore.
        ignore_suffixes: File suffixes to ignore.
        only_suffixes: Only include files with these suffixes.
        normalize: Whether to normalize paths.
        raw: Whether to include raw directory contents (ignores defaults).

    Yields:
        IncludedDataFile: Included data file objects found in the directory.
    """
    assert isRelativePath(dest_path), dest_path
    assert os.path.isdir(source_path), source_path

    ignore_dirs = tuple(ignore_dirs)
    if not raw:
        ignore_dirs += default_ignored_dirs

    ignore_filenames = tuple(ignore_filenames)
    if not raw:
        ignore_filenames += default_ignored_filenames

    ignore_suffixes = tuple(ignore_suffixes)
    if not raw:
        ignore_suffixes += default_ignored_suffixes

    for filename in getFileList(
        source_path,
        ignore_dirs=ignore_dirs,
        ignore_filenames=ignore_filenames,
        ignore_suffixes=ignore_suffixes,
        only_suffixes=only_suffixes,
        normalize=normalize,
    ):
        filename_relative = os.path.relpath(filename, source_path)

        filename_dest = os.path.join(dest_path, filename_relative)

        included_datafile = makeIncludedDataFile(
            source_path=filename,
            dest_path=filename_dest,
            reason=reason,
            tracer=tracer,
            tags=tags,
        )

        included_datafile.tags.add("raw-dir-contents" if raw else "data-dir-contents")

        yield included_datafile


def makeIncludedGeneratedDataFile(data, dest_path, reason, tracer, tags):
    """Create an included data file from generated data.

    Args:
        data: The binary data content.
        dest_path: The destination path in the distribution.
        reason: The reason for inclusion.
        tracer: The tracer to use for logging.
        tags: Tags associated with this file.

    Returns:
        IncludedDataFile: The created data file object.
    """
    assert isRelativePath(dest_path), dest_path

    if type(data) is str and str is not bytes:
        data = data.encode("utf8")

    assert type(data) is bytes, type(data)

    return IncludedDataFile(
        kind="data_blob",
        source_path=None,
        dest_path=dest_path,
        reason=reason,
        data=data,
        tracer=tracer,
        tags=tags,
    )


_included_data_files = []


def addIncludedDataFile(included_datafile):
    """Add a data file to the included list.

    Args:
        included_datafile: The IncludedDataFile object to add.
    """
    included_datafile.tags.update(getDataFileTags(included_datafile.dest_path))

    for external_datafile_pattern in getShallNotIncludeDataFilePatterns():
        if fnmatch.fnmatch(
            included_datafile.dest_path, external_datafile_pattern
        ) or isFilenameBelowPath(
            path=external_datafile_pattern, filename=included_datafile.dest_path
        ):
            included_datafile.tags.clear()
            included_datafile.tags.add("inhibit")

            return

    onDataFileTags(included_datafile)

    # TODO: Catch duplicates sooner.
    # for candidate in _included_data_files:
    #     if candidate.dest_path == included_datafile.dest_path:
    #         assert False, included_datafile

    if included_datafile.needsCopy():
        for external_datafile_pattern in getShallIncludeExternallyDataFilePatterns():
            if fnmatch.fnmatch(
                included_datafile.dest_path, external_datafile_pattern
            ) or isFilenameBelowPath(
                path=external_datafile_pattern, filename=included_datafile.dest_path
            ):
                included_datafile.tags.add("external")

                dest_path = getOutputPath(included_datafile.dest_path)

                if areSamePaths(dest_path, included_datafile.source_path):
                    inclusion_logger.sysexit(
                        """\
Error, when asking to copy files external data, you cannot output to\
same directory and need to use '--output-dir' option."""
                    )

    _included_data_files.append(included_datafile)


def getIncludedDataFiles():
    """Get all included data files.

    Returns:
        list: List of IncludedDataFile objects.
    """
    return _included_data_files


def checkProjectExpectedDataFiles():
    """Check if the expected data files are present.

    Notes:
        This is used to verify that the project configuration tools like
        'uv' or 'poetry' are not having their data files ignored by
        Nuitka or that Nuitka is not ignoring them.
    """

    expected_data_files = getProjectExpectedDataFiles()

    if expected_data_files is None:
        return

    expected_data_files = set(
        getNormalizedPath(filename) for filename in expected_data_files
    )

    included_data_files = set()

    for included_data_file in _included_data_files:
        if "user" in included_data_file.tags:
            included_data_files.add(included_data_file.dest_path)

    if expected_data_files != included_data_files:
        start_message = "Error, the expected data files from project configuration do not match the included data files."

        missing_data_files = expected_data_files - included_data_files
        extra_data_files = included_data_files - expected_data_files

        if missing_data_files:
            start_message += "\nMissing data files:\n" + "\n".join(
                sorted(missing_data_files)
            )

        if extra_data_files:
            start_message += "\nExtra data files:\n" + "\n".join(
                sorted(extra_data_files)
            )

        inclusion_logger.sysexit(start_message)


def _addIncludedDataFilesFromFileOptions():
    # Many different option variants, pylint: disable=too-many-branches

    for pattern, source_path, dest_path, arg in getShallIncludeDataFiles():
        filenames = resolveShellPatternToFilenames(pattern)

        count = 0
        for filename in filenames:
            file_reason = "specified data file '%s' on command line" % arg

            if source_path is None:
                rel_path = dest_path

                if rel_path.endswith(("/", os.path.sep)):
                    rel_path = os.path.join(rel_path, os.path.basename(filename))
            else:
                rel_path = os.path.join(dest_path, relpath(filename, source_path))

            rel_path = os.path.normpath(rel_path)

            if containsPathElements(rel_path, default_ignored_dirs):
                continue

            if os.path.basename(rel_path) in default_ignored_filenames:
                continue

            yield makeIncludedDataFile(
                filename, rel_path, file_reason, tracer=options_logger, tags="user"
            )

            count += 1

        if count == 0:
            options_logger.warning(
                "No matching data file to be included for '%s'." % pattern
            )

    for source_path, dest_path in getShallIncludeDataDirs():
        count = 0

        for included_datafile in makeIncludedDataDirectory(
            source_path=source_path,
            dest_path=os.path.normpath(dest_path),
            reason="specified data dir '%s' on command line" % source_path,
            tracer=options_logger,
            tags="user",
        ):
            yield included_datafile

            count += 1

        if count == 0:
            options_logger.warning("No data files in directory '%s'." % source_path)

    for source_path, dest_path in getShallIncludeRawDirs():
        count = 0

        for included_datafile in makeIncludedDataDirectory(
            source_path=source_path,
            dest_path=os.path.normpath(dest_path),
            reason="specified raw dir '%s' on command line" % source_path,
            tracer=options_logger,
            tags="user",
            raw=True,
        ):
            yield included_datafile

            count += 1

        if count == 0:
            options_logger.warning("No files in raw directory '%s.'" % source_path)


def addIncludedDataFilesFromFlavor():
    """Add data files required by the Python flavor/OS.

    Notes:
        Example: AIX requires libpython embedded.
    """
    if isAIX():
        # On AIX, the Python DLL is hidden in an archive.
        filename = "libpython%s.a" % python_version_str
        system_prefix = getSystemPrefixPath()

        if getArchitecture() == "64":
            lib_part = "lib64"
        else:
            return inclusion_logger.sysexit(
                """\
Error, change Nuitka code to define the path of the '%s' \
file in the Python installation '%s'."""
                % (filename, system_prefix)
            )

        filename_full = os.path.join(system_prefix, lib_part, filename)

        if not os.path.exists(filename_full):
            return inclusion_logger.sysexit(
                "Error, the defined path of the '%s' file in the Python installation '%s' is wrong."
                % (filename_full, system_prefix)
            )

        addIncludedDataFile(
            makeIncludedDataFile(
                source_path=filename_full,
                dest_path=filename,
                reason="Required Python DLL",
                tracer=inclusion_logger,
                tags="flavor",
            )
        )


def addIncludedDataFilesFromFileOptions():
    """Early data files, from user options that work with file system.

    Notes:
        These are processed early as they come from direct file/directory
        arguments.
    """

    for included_datafile in _addIncludedDataFilesFromFileOptions():
        addIncludedDataFile(included_datafile)


def scanIncludedPackageDataFiles(package_directory, pattern):
    """Scan a package directory for data files matching a pattern.

    Args:
        package_directory: The directory of the package to scan.
        pattern: The filename pattern to match.

    Returns:
        list: List of matching filenames.
    """
    ignore_suffixes = default_ignored_suffixes

    if pattern is not None:
        pattern_extension = getFilenameExtension(pattern)

        ignore_suffixes = tuple(
            ignore_suffix
            for ignore_suffix in ignore_suffixes
            if ignore_suffix != pattern_extension
        )

    result = []

    for pkg_filename in getFileList(
        package_directory,
        ignore_dirs=default_ignored_dirs,
        ignore_suffixes=ignore_suffixes,
        ignore_filenames=default_ignored_filenames,
    ):
        rel_path = os.path.relpath(pkg_filename, package_directory)

        if pattern and not fnmatch.fnmatch(rel_path, pattern):
            continue

        result.append(pkg_filename)

    return result


def makeIncludedPackageDataFiles(
    tracer, package_name, package_directory, pattern, reason, tags
):
    """Create included data files for a package.

    Args:
        tracer: The tracer to use for logging.
        package_name: The name of the package.
        package_directory: The directory of the package.
        pattern: The filename pattern to search for.
        reason: The reason for inclusion.
        tags: Tags associated with these files.

    Yields:
        IncludedDataFile: Included data file objects found.
    """
    tags = decodeDataFileTags(tags)
    tags.add("package_data")

    file_reason = "package '%s' %s" % (package_name, reason)

    for pkg_filename in scanIncludedPackageDataFiles(package_directory, pattern):
        rel_path = os.path.relpath(pkg_filename, package_directory)

        yield makeIncludedDataFile(
            source_path=pkg_filename,
            dest_path=os.path.join(package_name.asPath(), rel_path),
            reason=file_reason,
            tracer=tracer,
            tags=tags,
        )


def addIncludedDataFilesFromPlugins():
    """Add included data files from plugins.

    Notes:
        Plugins can yield included data files via the considerDataFiles hook.
    """
    # Plugins provide per module through this.
    for module in ModuleRegistry.getDoneModules():
        for included_datafile in considerDataFiles(module=module):
            addIncludedDataFile(included_datafile)


def addIncludedDataFilesFromPackageOptions():
    """Late data files, from plugins and user options that work with packages.

    Notes:
        These are processed after modules are located.
    """

    for package_name, filename_pattern in getShallIncludePackageData():
        package_name, package_directory, _module_kind, _finding = locateModule(
            module_name=package_name,
            parent_package=None,
            level=0,
        )

        if package_directory is None:
            options_logger.warning(
                "Failed to locate package directory of '%s'." % package_name.asString()
            )
            continue

        if os.path.isfile(package_directory):
            options_logger.sysexit(
                "Error, '%s' is a module not a package, cannot have package data."
                % package_name
            )

        # TODO: Maybe warn about packages that have no data files found.

        for included_datafile in makeIncludedPackageDataFiles(
            tracer=options_logger,
            package_name=package_name,
            package_directory=package_directory,
            pattern=filename_pattern,
            reason="package data",
            tags="user",
        ):
            addIncludedDataFile(included_datafile)


_data_file_traces = OrderedDict()


def _reportDataFiles():
    """Report (log) the included data files summary."""
    for key in _data_file_traces:
        count = len(_data_file_traces[key])
        tracer, reason = key

        # Avoid being too noisy, maybe add a control for increasing this.
        if count > 10:
            tracer.info("Included %d data files due to %s." % (count, reason))
        else:
            for kind, dest_path in _data_file_traces[key]:
                if kind == "data_blob":
                    tracer.info(
                        "Included data file '%s' due to %s."
                        % (
                            dest_path,
                            reason,
                        )
                    )
                elif kind == "data_file":
                    tracer.info(
                        "Included data file '%s' due to %s."
                        % (
                            dest_path,
                            reason,
                        )
                    )
                else:
                    assert False

    # Release the memory
    _data_file_traces.clear()


def _checkPathConflict(dest_path, standalone_entry_points):
    """Check for conflicts with standalone entry points.

    Args:
        dest_path: The destination path to check.
        standalone_entry_points: List of standalone entry points.
    """
    assert getNormalizedPath(dest_path) == dest_path

    while dest_path:
        for standalone_entry_point in standalone_entry_points:
            if dest_path == standalone_entry_point.dest_path:
                options_logger.sysexit(
                    """\
Error, data file to be placed in distribution as '%s' conflicts with %s '%s'."""
                    % (
                        dest_path,
                        standalone_entry_point.kind,
                        standalone_entry_point.dest_path,
                    )
                )
        dest_path = os.path.dirname(dest_path)


def _handleDataFile(included_datafile, standalone_entry_points):
    """Handle a single data file, writing or copying it.

    Args:
        included_datafile: The IncludedDataFile object.
        standalone_entry_points: List of standalone entry points to check conflicts against.

    Returns:
        tuple: (external, dest_path) where external is bool.
    """
    tracer = included_datafile.tracer

    if not isinstance(included_datafile, IncludedDataFile):
        tracer.sysexit("Error, can only accept 'IncludedData*' objects from ")

    if not isStandaloneMode():
        tracer.sysexit(
            "Error, package data files are only included in standalone or onefile mode."
        )

    key = tracer, included_datafile.reason
    if key not in _data_file_traces:
        _data_file_traces[key] = []

    _data_file_traces[key].append((included_datafile.kind, included_datafile.dest_path))

    dist_dir = getStandaloneDirectoryPath(bundle=True, real=False)

    if "external" in included_datafile.tags:
        dest_path = getOutputPath(included_datafile.dest_path)
        external = True
    else:
        _checkPathConflict(included_datafile.dest_path, standalone_entry_points)
        dest_path = os.path.join(dist_dir, included_datafile.dest_path)
        external = False

    if included_datafile.kind == "data_blob":
        makePath(os.path.dirname(dest_path))

        with openTextFile(filename=dest_path, mode="wb") as output_file:
            output_file.write(included_datafile.data)

        if "script" in included_datafile.tags:
            addFileExecutablePermission(dest_path)

    elif included_datafile.kind == "data_file":
        makePath(os.path.dirname(dest_path))

        copyFileWithPermissions(
            source_path=included_datafile.source_path,
            dest_path=dest_path,
            target_dir=dist_dir,
        )
    else:
        assert False, included_datafile

    return external, dest_path


def copyDataFiles(standalone_entry_points):
    """Copy the data files needed for standalone distribution.

    Notes:
        This is for data files only, not DLLs or even extension modules,
        those must be registered as entry points, and would not go through
        necessary handling if provided like this.

    Args:
        standalone_entry_points: List of standalone entry points available.

    Returns:
        list: List of paths to the copied data files.
    """

    data_file_paths = []

    for included_datafile in getIncludedDataFiles():
        # TODO: directories should be resolved to files.
        if included_datafile.needsCopy():
            if shallMakeModule():
                options_logger.sysexit(
                    """\
Error, data files for modules must be done via wheels, or commercial plugins \
'--embed-*' options. Not done for '%s'."""
                    % included_datafile.dest_path
                )
            elif not isStandaloneMode():
                options_logger.sysexit(
                    """\
Error, data files cannot be included in accelerated mode unless using commercial \
plugins '--embed-*' options. Not done for '%s'."""
                    % included_datafile.dest_path
                )

            external, data_file_path = _handleDataFile(
                included_datafile=included_datafile,
                standalone_entry_points=standalone_entry_points,
            )

            if not external:
                data_file_paths.append(data_file_path)

    _reportDataFiles()

    return data_file_paths


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""
Tool to extract build configuration from uv_build projects.

This script executes the PEP 517 hook `prepare_metadata_for_build_wheel`
to generate the metadata directory, then parses it to find entry points
and package information. It outputs Nuitka arguments directly.
"""

import csv
import json
import os
import sys
import tempfile
import zipfile

from nuitka.containers.OrderedDicts import OrderedDict

# We expect `uv_build` to be installed in the environment running this script.
try:
    import uv_build
except ImportError:
    sys.exit("Error, 'uv_build' not installed in this environment.")


# uv_build requires 'uv-build' or 'uv' executable to be in PATH.
# On Windows, it's often in Scripts/ but not in PATH during tests or some environments.
# In a venv, python is in Scripts/ folder, so that is the one to add.
candidates = (
    os.path.dirname(sys.executable),
    os.path.join(os.path.dirname(sys.executable), "Scripts"),
)

for script_dir in candidates:
    if os.path.exists(script_dir) and script_dir not in os.environ["PATH"]:
        os.environ["PATH"] = script_dir + os.pathsep + os.environ["PATH"]


def _parseMetaData(metadata_file):
    info = {"Name": "", "Requires-Dist": []}
    if os.path.exists(metadata_file):
        with open(metadata_file, "r", encoding="utf-8") as f:
            for line in f:
                if line.startswith("Name: "):
                    info["Name"] = line.split(": ", 1)[1].strip()
                elif line.startswith("Requires-Dist: "):
                    info["Requires-Dist"].append(line.split(": ", 1)[1].strip())
    return info


def _parseEntryPoints(entry_points_file):
    scripts = OrderedDict()

    current_section = None
    with open(entry_points_file, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            if line.startswith("[") and line.endswith("]"):
                current_section = line[1:-1]
                continue

            if current_section in ("console_scripts", "gui_scripts"):
                if "=" in line:
                    name, value = line.split("=", 1)
                    scripts[name.strip()] = value.strip()
    return scripts


def _extractFromUvBuild(temp_dir):
    # This is extremely detail rich, pylint: disable=too-many-branches,too-many-locals

    arguments = []

    # Generate metadata by building the wheel
    # We need to build the wheel because prepare_metadata_for_build_wheel
    # does not generate a complete RECORD file with all package files.
    wheel_filename = uv_build.build_wheel(temp_dir)
    wheel_path = os.path.join(temp_dir, wheel_filename)

    dist_info_dir = None

    with zipfile.ZipFile(wheel_path) as zf:
        for name in zf.namelist():
            if name.endswith(".dist-info/RECORD"):
                dist_info_dir = os.path.dirname(name)
                break

        if not dist_info_dir:
            sys.exit("Error, failed to find .dist-info directory in generated wheel.")

        # Extract the .dist-info directory to temp_dir so existing parsing logic works
        for name in zf.namelist():
            if name.startswith(dist_info_dir):
                zf.extract(name, temp_dir)

    metadata_path = os.path.join(temp_dir, dist_info_dir)

    # Parse METADATA for Name and Dependencies
    metadata_file = os.path.join(metadata_path, "METADATA")
    pkg_info = _parseMetaData(metadata_file)

    project_name = pkg_info.get("Name")
    requirements = pkg_info.get("Requires-Dist", [])
    for req in requirements:
        arguments.append("--pyproject-requires=%s" % req)

    # Parse entry points
    entry_points_file = os.path.join(metadata_path, "entry_points.txt")
    if os.path.exists(entry_points_file):
        scripts = _parseEntryPoints(entry_points_file)

        for binary_name, entry_point in scripts.items():
            module_name, function_name = entry_point.split(":", 1)

            arguments.append(
                "--main-entry-point=%s=%s:%s"
                % (binary_name, module_name, function_name)
            )
            arguments.append("--output-filename=%s" % binary_name)
            arguments.append("--output-folder-name=%s" % binary_name)

    # Parse RECORD for packages and data files
    record_file = os.path.join(metadata_path, "RECORD")
    packages = set()
    data_files = []

    if os.path.exists(record_file):

        with open(record_file, "r", encoding="utf-8") as f:
            reader = csv.reader(f)
            for row in reader:
                path = row[0]

                # Skip the RECORD file itself and .dist-info directory content
                if path == "RECORD" or path.startswith(dist_info_dir):
                    continue

                parts = path.split("/")

                # Check for Python files or extension modules to identify packages
                if path.endswith((".py", ".pyd", ".so")):
                    # The top-level folder or file (module) is what we care about
                    packages.add(parts[0].replace(".py", ""))
                else:
                    data_files.append(path)

    return {
        "arguments": arguments,
        "project_name": project_name,
        "packages": list(packages),
        "data_files": data_files,
    }


def main():
    if len(sys.argv) != 2:
        sys.exit("Usage: %s <dump_filename>" % sys.argv[0])

    dump_filename = sys.argv[1]

    with tempfile.TemporaryDirectory() as temp_dir:
        result = _extractFromUvBuild(temp_dir)

        with open(dump_filename, "w", encoding="utf-8") as f:
            json.dump(result, f, indent=4)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Expression codes, side effects, or statements that are an unused expression.

When you write "f()", i.e. you don't use the return value, that is an expression
only statement.

"""

from .CodeHelpers import generateExpressionCode
from .ErrorCodes import getReleaseCode


def generateExpressionOnlyCode(statement, emit, context):
    return getStatementOnlyCode(
        value=statement.subnode_expression, emit=emit, context=context
    )


def getStatementOnlyCode(value, emit, context):
    tmp_name = context.allocateTempName(
        base_name="unused", type_name="nuitka_void", unique=True
    )
    tmp_name.maybe_unused = True

    generateExpressionCode(
        expression=value, to_name=tmp_name, emit=emit, context=context
    )

    # An error of the expression is dealt inside of this, not necessary here,
    # but we have to release non-error value if it has a reference.
    getReleaseCode(release_name=tmp_name, emit=emit, context=context)


def generateSideEffectsCode(to_name, expression, emit, context):
    for side_effect in expression.subnode_side_effects:
        getStatementOnlyCode(value=side_effect, emit=emit, context=context)

    generateExpressionCode(
        to_name=to_name,
        expression=expression.subnode_expression,
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Progress bars in Nuitka.

This is responsible for wrapping the rendering of progress bar and emitting tracing
to the user while it's being displayed.

"""

import atexit
import sys
from contextlib import contextmanager

from nuitka import Tracing
from nuitka.Tracing import general, getDisableStylesCode
from nuitka.utils.Importing import importFromInlineCopy
from nuitka.utils.ThreadedExecutor import RLock
from nuitka.utils.Utils import isWin32Windows, withNoExceptions

# spell-checker: ignore tqdm,ncols

# Late import and optional to be there.
use_progress_bar = None
_tqdm = None
_colorama = None


def wrapWithStyles(value, styles):
    if use_progress_bar == "tqdm":
        return Tracing.wrapWithStyles(value, styles)
    else:
        return "[%s]%s" % (" ".join(styles), value)


_uses_threading = False

_progress_info_style = ("bold", "blue")
_progress_percentage_style = ("bold", "green")


def enableThreading():
    """Inform about threading being used."""

    # Singleton, pylint: disable=global-statement
    global _uses_threading

    if not _uses_threading:
        # Install the redirection for stdout/stderr to ensure that partial lines
        # which are common for threaded tools do not get mixed with progress bar.

        if not hasattr(sys.stdout, "original_stream"):
            sys.stdout = Tracing.LineBufferedStreamRedirector(sys.stdout)

        if not hasattr(sys.stderr, "original_stream"):
            sys.stderr = Tracing.LineBufferedStreamRedirector(sys.stderr)

        _uses_threading = True


class NuitkaProgressBarTqdm(object):
    def __init__(self, iterable, stage, total, min_total, unit):
        self.total = total

        # The minimum may not be provided, then default to 0.
        self.min_total = min_total

        # No item under work yet.
        self.item = None

        # No progress yet.
        self.progress = 0

        bar_format = "%s%s|{bar:25}| {n_fmt}/{total_fmt}{unit}%s%s" % (
            wrapWithStyles("{desc}", styles=_progress_info_style),
            wrapWithStyles("{percentage:5.1f}%", styles=_progress_percentage_style),
            wrapWithStyles("{postfix}", styles=_progress_info_style),
            getDisableStylesCode(),
        )

        # Ensure we use the original stream for tqdm, so it doesn't trigger our redirection
        # which would cause infinite recursion.
        tqdm_file = sys.stderr
        if hasattr(tqdm_file, "original_stream"):
            tqdm_file = tqdm_file.original_stream

        # Render immediately with 0 progress, and setting disable=None enables tty detection.
        self.tqdm = _tqdm(
            iterable=iterable,
            initial=self.progress,
            total=(
                max(self.total, self.min_total) if self.min_total is not None else None
            ),
            unit=(" " + unit + "s") if unit else "",
            disable=None,
            leave=False,
            dynamic_ncols=True,
            bar_format=bar_format,
            file=tqdm_file,
        )

        self.tqdm.set_description(stage)
        self.setCurrent(self.item)

    def __iter__(self):
        return iter(self.tqdm)

    def updateTotal(self, total):
        if total != self.total:
            self.total = total
            self.tqdm.total = max(total, self.min_total)

    def setCurrent(self, item):
        if item != self.item:
            self.item = item

            if item is not None:
                self.tqdm.set_postfix_str(item)
            else:
                self.tqdm.set_postfix()

    def update(self):
        self.progress += 1
        self.tqdm.update(1)

    def clear(self):
        self.tqdm.clear()

    def close(self):
        self.tqdm.close()

    @contextmanager
    def withExternalWritingPause(self):
        # We perform what "external_write_mode" does manually, so we can force the
        # refresh to happen, which it seemingly doesn't do reliably.

        # spell-checker: ignore nolock
        if _uses_threading:
            lock = self.tqdm.get_lock()
            lock.acquire()

        try:
            self.tqdm.clear(nolock=True)
            yield
            self.tqdm.refresh(nolock=True)
        finally:
            if _uses_threading:
                lock.release()


class NuitkaProgressBarRich(object):
    def __init__(self, iterable, stage, total, min_total, unit):
        self.iterable = iterable or ()

        if total is None and hasattr(iterable, "__len__"):
            self.total = len(iterable)
        else:
            self.total = total

        self.min_total = min_total
        self.item = None

        class TextColumnCropped(_rich_progress.TextColumn):
            def __init__(
                self,
                text_format,
                style="none",
                justify="left",
                markup=True,
                highlighter=None,
                table_column=None,
            ):
                # False alarm, pylint: disable=non-parent-init-called
                _rich_progress.TextColumn.__init__(
                    self,
                    text_format=text_format,
                    style=style,
                    justify=justify,
                    markup=markup,
                    highlighter=highlighter,
                    table_column=table_column,
                )

            def render(self, task):
                _text = self.text_format.format(task=task)
                if self.markup:
                    text = _rich_progress.Text.from_markup(
                        _text, style=self.style, justify=self.justify, overflow="crop"
                    )
                else:
                    text = _rich_progress.Text(
                        _text, style=self.style, justify=self.justify, overflow="crop"
                    )
                if self.highlighter:
                    self.highlighter.highlight(text)
                return text

        # Unwrap our own redirection for Scons.
        rich_file = sys.stdout
        if hasattr(rich_file, "original_stream"):
            rich_file = rich_file.original_stream

        # This is for "rich" to not use the "SconsStreamRedirector" that we install
        # for Scons, but instead use the one that really goes to the terminal.
        console = _rich_console.Console(file=rich_file)

        self.rich_progress = _rich_progress.Progress(
            _rich_progress.TextColumn(
                wrapWithStyles("{task.description}", styles=_progress_info_style),
                justify="full",
            ),
            _rich_progress.BarColumn(bar_width=25),
            _rich_progress.TextColumn(
                wrapWithStyles(
                    "{task.percentage:>5.1f}%", styles=_progress_percentage_style
                ),
                justify="full",
            ),
            "|",
            _rich_progress.TextColumn(
                "{task.completed:>0.0f}/{task.total:>0.0f}{task.fields[unit_label]}",
                justify="full",
            ),
            _rich_progress.TextColumn("{task.fields[postfix_bullet]}"),
            TextColumnCropped(
                wrapWithStyles("{task.fields[postfix]}", styles=_progress_info_style),
            ),
            refresh_per_second=10000,
            transient=True,
            redirect_stdout=False,
            redirect_stderr=False,
            console=console,
        )
        self.rich_progress.start()

        effective_total = self.total
        if self.total is not None:
            if self.min_total is not None:
                effective_total = max(self.total, self.min_total)
            else:
                effective_total = max(self.total, 0)
        elif self.min_total is not None and self.min_total > 0:
            effective_total = self.min_total
        else:
            effective_total = None

        self.task_id = self.rich_progress.add_task(
            description=stage,
            total=effective_total,
            postfix="",
            postfix_bullet="",
            unit_label=(" " + unit + "s") if unit else "",
            start=True,
        )
        self.setCurrent(self.item)

    def __iter__(self):
        self.rich_progress.start_task(self.task_id)

        for val in self.iterable:
            yield val
            self.update()

    def updateTotal(self, total):
        if total != self.total:
            self.total = total
            effective_total = total
            if self.min_total is not None:
                effective_total = max(total, self.min_total)
            self.rich_progress.update(self.task_id, total=effective_total)

    def setCurrent(self, item):
        if item != self.item:
            self.item = item

            self.rich_progress.update(
                self.task_id,
                postfix=str(item),
                postfix_bullet="|" if item is not None else "",
            )

    def update(self):
        self.rich_progress.update(self.task_id, advance=1)

    def clear(self):
        # Rich progress with transient=True clears on stop, so this is not
        # needed.
        pass

    def close(self):
        if self.rich_progress.live.is_started:
            # Ensure task is marked as finished if not already
            if (
                self.task_id in self.rich_progress.task_ids
                and not self.rich_progress.tasks[
                    self.rich_progress.task_ids.index(self.task_id)
                ].finished
            ):

                with withNoExceptions():
                    self.rich_progress.stop_task(self.task_id)

            with withNoExceptions():
                self.rich_progress.stop()

    @contextmanager
    def withExternalWritingPause(self):
        # Temporarily stop (and clear due to transient=True) the Rich progress display
        # to allow external printing, then restart it.
        was_started = self.rich_progress.live.is_started

        if was_started:
            self.rich_progress.stop()

        try:
            yield
        finally:
            if was_started:
                self.rich_progress.start()


def _getTqdmModule():
    """Get the tqdm module if possible, might return None."""
    global _tqdm  # singleton, pylint: disable=global-statement

    if _tqdm:
        return _tqdm
    elif _tqdm is False:
        return None
    else:
        _tqdm = importFromInlineCopy("tqdm", must_exist=False, delete_module=True)

        if _tqdm is None:
            try:
                # Cannot use import tqdm due to pylint bug.
                import tqdm as tqdm_installed  # pylint: disable=I0021,import-error

                _tqdm = tqdm_installed
            except ImportError:
                # We handle the case without inline copy too, but it may be removed, e.g. on
                # Debian it's only a recommended install, and not included that way.
                pass

        if _tqdm is None:
            _tqdm = False
            return None

        _tqdm = _tqdm.tqdm

        # Tolerate the absence ignore the progress bar
        _tqdm.set_lock(RLock())

        return _tqdm


# Global variable to cache the rich module instance or import failure
_rich_progress = None
_rich_console = None


def _getRichModule():
    """Get the rich module if possible, might return None."""
    # Singleton, pylint: disable=global-statement
    global _rich_progress, _rich_console

    if _rich_progress:
        return _rich_progress
    elif _rich_progress is False:
        return None

    if _rich_progress is None:
        try:
            # Try importing from pip's vendored packages first
            import pip._vendor.rich.console as vendored_rich_console
            import pip._vendor.rich.progress as vendored_rich_progress

            _rich_progress = vendored_rich_progress
            _rich_console = vendored_rich_console
        except ImportError:
            try:
                import rich.console as standard_rich_console
                import rich.progress as standard_rich_progress

                _rich_progress = standard_rich_progress
                _rich_console = standard_rich_console
            except ImportError:
                _rich_progress = False

        if _rich_progress:
            if not hasattr(_rich_progress, "Progress"):
                _rich_progress = False

    if _rich_progress is False:
        return None
    else:
        atexit.register(_cleanupRich)

        return _rich_progress


def _checkRichModule():
    if _getRichModule() is not None:
        return "rich"
    else:
        return "none"


def _checkTqdmModule():
    global _colorama  # singleton, pylint: disable=global-statement

    if _getTqdmModule() is not None:
        if isWin32Windows():
            if _colorama is None:
                _colorama = importFromInlineCopy(
                    "colorama", must_exist=True, delete_module=True
                )

            # Check for our redirection, and temporarily disable it, so colorama
            # finds the real handles.
            if hasattr(sys.stdout, "original_stream"):
                stdout_redirector = sys.stdout
                sys.stdout = stdout_redirector.original_stream
            else:
                stdout_redirector = None

            if hasattr(sys.stderr, "original_stream"):
                stderr_redirector = sys.stderr
                sys.stderr = stderr_redirector.original_stream
            else:
                stderr_redirector = None

            _colorama.init()

            if stdout_redirector:
                stdout_redirector.original_stream = sys.stdout
                sys.stdout = stdout_redirector

            if stderr_redirector:
                stderr_redirector.original_stream = sys.stderr
                sys.stderr = stderr_redirector

        return "tqdm"
    else:
        return "none"


# Try progress bars in this order.
_default_progress_bars = ("tqdm", "rich")


def enableProgressBar(progress_bar):
    global use_progress_bar  # singleton, pylint: disable=global-statement

    if use_progress_bar is None:
        use_progress_bar = "none"

        if progress_bar == "auto":
            check_progress_bars = _default_progress_bars
        elif progress_bar in _default_progress_bars:
            check_progress_bars = (progress_bar,)
        else:
            check_progress_bars = ()

        for progress_bar_check in check_progress_bars:
            if progress_bar_check == "rich":
                use_progress_bar = _checkRichModule()
            elif progress_bar_check == "tqdm":
                use_progress_bar = _checkTqdmModule()
            else:
                assert False, progress_bar_check

            if use_progress_bar != "none":
                break


def _cleanupRich():
    if use_progress_bar == "rich" and Tracing.progress is not None:
        with withNoExceptions():
            Tracing.progress.close()


def setupProgressBar(stage, unit, total, min_total=0):
    # Make sure the other was closed.
    assert Tracing.progress is None

    if use_progress_bar == "rich":
        Tracing.progress = NuitkaProgressBarRich(
            iterable=None,
            stage=stage,
            total=total,
            min_total=min_total,
            unit=unit,
        )
    elif use_progress_bar == "tqdm":
        Tracing.progress = NuitkaProgressBarTqdm(
            iterable=None,
            stage=stage,
            total=total,
            min_total=min_total,
            unit=unit,
        )


def reportProgressBar(item, total=None, update=True):
    if Tracing.progress is not None:
        try:
            if total is not None:
                Tracing.progress.updateTotal(total)

            Tracing.progress.setCurrent(item)

            if update:
                Tracing.progress.update()
        except Exception as e:  # Catch all the things, pylint: disable=broad-except
            # We disable the progress bar now, because it's causing issues.
            general.warning("Progress bar disabled due to bug: %s" % (str(e)))
            closeProgressBar()


def closeProgressBar():
    """Close the active progress bar.

    Returns: int or None - if displayed, the total used last time.
    """

    if Tracing.progress is not None:
        # Retrieve that previous total, for repeated progress bars, it
        # can be used as a new minimum.
        result = Tracing.progress.total

        Tracing.progress.close()
        Tracing.progress = None

        if sys.stdout.isatty():
            Tracing.my_print(Tracing.getDisableStylesCode(), end="")

        return result


def wrapWithProgressBar(iterable, stage, unit):
    if use_progress_bar == "rich":
        result = NuitkaProgressBarRich(
            iterable=iterable, unit=unit, stage=stage, total=None, min_total=None
        )
        Tracing.progress = result
        return result
    elif use_progress_bar == "tqdm":
        result = NuitkaProgressBarTqdm(
            iterable=iterable, unit=unit, stage=stage, total=None, min_total=None
        )
        Tracing.progress = result
        return result
    else:
        return iterable


@contextmanager
def withNuitkaDownloadProgressBar(*args, **kwargs):
    if use_progress_bar in (None, "none"):
        yield None
        return

    # Check if stdout is a TTY for Rich
    is_tty = sys.stdout.isatty()

    if use_progress_bar == "rich":
        description = kwargs.get("desc", "Downloading")
        total_size_bytes = kwargs.get("total", 0)

        # Unwrap our own redirection if we are active during SCons or similar.
        rich_file = sys.stdout
        if hasattr(rich_file, "original_stream"):
            rich_file = rich_file.original_stream

        console = _rich_console.Console(file=rich_file)

        _rich_progress_cm = _rich_progress.Progress(
            _rich_progress.TextColumn("[bold blue]{task.description}", justify="right"),
            _rich_progress.BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "\u2022",
            _rich_progress.DownloadColumn(),
            "\u2022",
            _rich_progress.TransferSpeedColumn(),
            "\u2022",
            _rich_progress.TimeRemainingColumn(),
            transient=True,
            disable=not is_tty,
            console=console,
        )

        with _rich_progress_cm as rich_p_instance:
            task_id = rich_p_instance.add_task(
                description, total=total_size_bytes or None
            )

            def onProgressRich(block_num=1, block_size=1, total_size=None):
                # pylint: disable=unused-argument

                if total_size is not None:
                    rich_p_instance.update(task_id, total=total_size)

            yield onProgressRich
    else:

        class NuitkaDownloadProgressBarTqdm(_tqdm):
            # spell-checker: ignore bsize, tsize
            def onProgress(self, b=1, bsize=1, tsize=None):
                if tsize is not None:
                    self.total = tsize  # false alarm, pylint: disable=I0021,attribute-defined-outside-init
                self.update(b * bsize - self.n)

        tqdm_kwargs = kwargs.copy()
        tqdm_kwargs.update(
            disable=None,
            leave=False,
            dynamic_ncols=True,
            bar_format="{desc} {percentage:3.1f}%|{bar:25}| {n_fmt}/{total_fmt}{postfix}"
            + getDisableStylesCode(),
        )
        with NuitkaDownloadProgressBarTqdm(*args, **tqdm_kwargs) as pb_tqdm:
            yield pb_tqdm.onProgress


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Shared definitions of what comparison operation helpers are available.

These are functions to work with helper names, as well as sets of functions to
generate specialized code variants for.

Note: These are ordered, so we can define the order they are created in the
code generation of specialized helpers, as this set is used for input there
too.

"""

from nuitka.containers.OrderedSets import buildOrderedSet

# Mapping of rich comparison Python level name, to C level name of helpers.
rich_comparison_codes = {
    "Lt": "LT",
    "LtE": "LE",
    "Eq": "EQ",
    "NotEq": "NE",
    "Gt": "GT",
    "GtE": "GE",
}

# Subset of comparisons which will be used for identical types.
rich_comparison_subset_codes = {
    "Lt": "LT",
    "LtE": "LE",
    "Eq": "EQ",
}


def _makeDefaultOps():
    for comparator in rich_comparison_codes.values():
        yield "RICH_COMPARE_%s_OBJECT_OBJECT_OBJECT" % comparator
        yield "RICH_COMPARE_%s_NBOOL_OBJECT_OBJECT" % comparator


def _makeTypeOps(type_name, may_raise_same_type, shortcut=False):
    for result_part in "OBJECT", "CBOOL", "NBOOL":
        for comparator in rich_comparison_codes.values():
            if result_part == "CBOOL":
                continue

            # TODO: We want those as well, but we don't need them immediately.
            if type_name in ("NILONG",):
                continue

            yield "RICH_COMPARE_%s_%s_OBJECT_%s" % (comparator, result_part, type_name)
            yield "RICH_COMPARE_%s_%s_%s_OBJECT" % (comparator, result_part, type_name)

        if may_raise_same_type and result_part == "CBOOL":
            continue
        if not may_raise_same_type and result_part == "NBOOL":
            continue

        for comparator in (
            rich_comparison_codes.values()
            if not shortcut
            else rich_comparison_subset_codes.values()
        ):
            yield "RICH_COMPARE_%s_%s_%s_%s" % (
                comparator,
                result_part,
                type_name,
                type_name,
            )


def _makeFriendOps(type_name1, type_name2, may_raise, shortcut):
    assert type_name1 != type_name2

    for result_part in "OBJECT", "CBOOL", "NBOOL":
        if not may_raise:
            if result_part == "NBOOL":
                continue

        for comparator in (
            rich_comparison_codes.values()
            if not shortcut
            else rich_comparison_subset_codes.values()
        ):
            yield "RICH_COMPARE_%s_%s_%s_%s" % (
                comparator,
                result_part,
                type_name1,
                type_name2,
            )


specialized_cmp_helpers_set = buildOrderedSet(
    # Default implementation.
    _makeDefaultOps(),
    # Pure types:
    _makeTypeOps("STR", may_raise_same_type=False, shortcut=True),
    _makeTypeOps("UNICODE", may_raise_same_type=False, shortcut=True),
    _makeTypeOps("BYTES", may_raise_same_type=False, shortcut=True),
    _makeTypeOps("INT", may_raise_same_type=False, shortcut=True),
    _makeTypeOps("LONG", may_raise_same_type=False, shortcut=True),
    _makeTypeOps("FLOAT", may_raise_same_type=False, shortcut=True),
    # Dual types
    _makeTypeOps("NILONG", may_raise_same_type=False, shortcut=True),
    # TODO: What would shortcut mean, how do tuples compare their elements then?
    _makeTypeOps("TUPLE", may_raise_same_type=True, shortcut=False),
    _makeTypeOps("LIST", may_raise_same_type=True, shortcut=False),
    # Mixed Python types:
    # TODO: Absolutely possible to shortcut, why aren't we doing it?
    _makeFriendOps("LONG", "INT", may_raise=False, shortcut=False),
    # Partial Python with C types
    # TODO: Absolutely possible to shortcut, why aren't we doing it?
    _makeFriendOps("INT", "CLONG", may_raise=False, shortcut=False),
    _makeFriendOps("LONG", "DIGIT", may_raise=False, shortcut=False),
    _makeFriendOps("FLOAT", "CFLOAT", may_raise=False, shortcut=False),
    # Partial dual types with C types, cannot shortcut as reverse argument versions are used.
    _makeFriendOps("NILONG", "CLONG", may_raise=False, shortcut=False),
    _makeFriendOps("NILONG", "DIGIT", may_raise=False, shortcut=False),
    # TODO: Add "CLONG_CLONG" type ops once we use that for local variables too.
)

_non_specialized_cmp_helpers_set = set()


def getSpecializedComparisonOperations():
    return specialized_cmp_helpers_set


def getNonSpecializedComparisonOperations():
    return _non_specialized_cmp_helpers_set


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Display the DLLs in a package."""

import os

from nuitka.freezer.DllDependenciesCommon import (
    getPackageSpecificDLLDirectories,
)
from nuitka.importing.Importing import (
    addMainScriptDirectory,
    hasMainScriptDirectory,
    locateModule,
)
from nuitka.Tracing import tools_logger
from nuitka.tree.SourceHandling import readSourceCodeFromFilename
from nuitka.utils.FileOperations import (
    listDllFilesFromDirectory,
    listExeFilesFromDirectory,
    relpath,
)
from nuitka.utils.Importing import (
    getExtensionModuleSuffixes,
    getPackageDirFilename,
)
from nuitka.utils.ModuleNames import ModuleName
from nuitka.utils.SharedLibraries import getDllExportedSymbols
from nuitka.utils.Utils import isMacOS


def getPythonEntryPointExportedSymbolName(module_name):
    result = "%s%s" % ("init" if str is bytes else "PyInit_", module_name.asString())

    if isMacOS():
        result = "_" + result

    return result


def isFileExtensionModule(module_filename):
    for suffix in getExtensionModuleSuffixes():
        if module_filename.endswith(suffix):
            module_name = ModuleName(os.path.basename(module_filename)[: -len(suffix)])

            exported_symbols = getDllExportedSymbols(
                logger=tools_logger, filename=module_filename
            )

            if exported_symbols is None:
                return None

            return (
                getPythonEntryPointExportedSymbolName(module_name) in exported_symbols
            )

    return False


def scanModule(module_name, scan_function):
    module_name = ModuleName(module_name)

    if not hasMainScriptDirectory():
        addMainScriptDirectory(os.getcwd())

    _module_name, package_directory, module_kind, finding = locateModule(
        module_name=module_name, parent_package=None, level=0
    )

    if finding == "not-found":
        tools_logger.sysexit(
            "Error, cannot find '%s' package." % module_name.asString()
        )

    if not os.path.isdir(package_directory):
        tools_logger.sysexit(
            "Error, doesn't seem that '%s' is a package on disk."
            % module_name.asString()
        )

    from nuitka.plugins.Plugins import activatePlugins

    activatePlugins()

    if module_kind != "extension":
        package_filename = getPackageDirFilename(package_directory)

        if package_filename is not None:
            readSourceCodeFromFilename(module_name, package_filename, pre_load=False)

    tools_logger.info(
        "Checking package '%s' directory '%s' .. " % (module_name, package_directory)
    )

    for package_dll_dir in getPackageSpecificDLLDirectories(
        module_name, consider_plugins=True
    ):
        for package_dll_filename, _dll_basename in scan_function(package_dll_dir):
            if isFileExtensionModule(package_dll_filename):
                continue

            yield package_directory, package_dll_filename, package_dll_dir


def displayDLLs(module_name):
    """Display the DLLs for a module name."""

    count = 0

    for package_directory, package_dll_filename, _package_dll_dir in scanModule(
        module_name=module_name, scan_function=listDllFilesFromDirectory
    ):
        tools_logger.my_print(
            "  %s" % relpath(package_dll_filename, start=package_directory),
        )

        count += 1

    tools_logger.info("Found %s DLLs." % count)


def displayEXEs(module_name):
    """Display the EXEs for a module name."""

    count = 0

    for package_directory, package_dll_filename, _package_dll_dir in scanModule(
        module_name=module_name, scan_function=listExeFilesFromDirectory
    ):

        tools_logger.my_print(
            "  %s" % relpath(package_dll_filename, start=package_directory),
        )

        count += 1

    tools_logger.info("Found %s EXEs." % count)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



__version__="4.10.1"
__copyright__="Copyright (c) 2001 - 2025 The SCons Foundation"
__developer__="bdbaddog"
__date__="Sun, 16 Nov 2025 14:18:20 -0700"
__buildsys__="M1Dog2021"
__revision__="055b01f429d58b686701a56df863a817c36bb103"
__build__="055b01f429d58b686701a56df863a817c36bb103"
# make sure compatibility is always in place
import SCons.compat  # noqa


#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""
This module provides common functionality for build package backends.

It stores global state about the project that is being built, e.g. the
project name and the expected data files.
"""

_project_name = None
_project_expected_data_files = None


def getProjectName():
    return _project_name


def setProjectName(project_name):
    global _project_name  # singleton, pylint: disable=global-statement
    _project_name = project_name


def getProjectExpectedDataFiles():
    return _project_expected_data_files


def setProjectExpectedDataFiles(project_expected_data_files):
    global _project_expected_data_files  # singleton, pylint: disable=global-statement
    _project_expected_data_files = project_expected_data_files


def reportBuildError(logger, name, command, stdout, stderr):
    if str is not bytes:
        if isinstance(stdout, bytes):
            stdout = stdout.decode("utf-8", "replace")

        if isinstance(stderr, bytes):
            stderr = stderr.decode("utf-8", "replace")

    logger.sysexit(
        """\
Error, failed to extract build configuration via '%(name)s'.
Error output from '%(name)s' configuration extraction:
Command used: %(command)s
stdout: %(stdout)s
stderr: %(stderr)s"""
        % {
            "name": name,
            "command": command,
            "stdout": stdout,
            "stderr": stderr,
        }
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Subscript related code generation.

There is special handling for integer indexes, which can be dealt with
much faster than general subscript lookups.
"""

from nuitka.States import states

from .CodeHelpers import (
    generateChildExpressionCode,
    generateExpressionCode,
    generateExpressionsCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitBoolCode, getErrorExitCode, getReleaseCodes


def decideIntegerSubscript(subscript):
    if subscript.isExpressionConstantRef():
        if subscript.isIndexConstant():
            constant_value = subscript.getIndexValue()

            if abs(constant_value) < 2**31:
                return constant_value, True

    return None, False


def generateAssignmentSubscriptCode(statement, emit, context):
    subscribed = statement.subnode_subscribed
    subscript = statement.subnode_subscript
    value = statement.subnode_source

    subscript_constant, integer_subscript = decideIntegerSubscript(subscript)

    value_name = context.allocateTempName("ass_subvalue")

    generateExpressionCode(
        to_name=value_name, expression=value, emit=emit, context=context
    )

    subscribed_name = context.allocateTempName("ass_subscribed")
    generateExpressionCode(
        to_name=subscribed_name, expression=subscribed, emit=emit, context=context
    )

    subscript_name = context.allocateTempName("ass_subscript")
    generateExpressionCode(
        to_name=subscript_name, expression=subscript, emit=emit, context=context
    )

    with context.withCurrentSourceCodeReference(
        value.getSourceReference()
        if states.is_full_compat
        else statement.getSourceReference()
    ):
        if integer_subscript:
            _getIntegerSubscriptAssignmentCode(
                subscribed_name=subscribed_name,
                subscript_name=subscript_name,
                subscript_value=subscript_constant,
                value_name=value_name,
                emit=emit,
                context=context,
            )
        else:
            _getSubscriptAssignmentCode(
                target_name=subscribed_name,
                subscript_name=subscript_name,
                value_name=value_name,
                emit=emit,
                context=context,
            )


def generateDelSubscriptCode(statement, emit, context):
    subscribed = statement.subnode_subscribed
    subscript = statement.subnode_subscript

    target_name, subscript_name = generateExpressionsCode(
        expressions=(subscribed, subscript),
        names=("delsubscr_target", "delsubscr_subscript"),
        emit=emit,
        context=context,
    )

    with context.withCurrentSourceCodeReference(
        subscript.getSourceReference()
        if states.is_full_compat
        else statement.getSourceReference()
    ):
        _getSubscriptDelCode(
            target_name=target_name,
            subscript_name=subscript_name,
            emit=emit,
            context=context,
        )


def generateSubscriptLookupCode(to_name, expression, emit, context):
    subscribed = expression.subnode_expression
    subscript = expression.subnode_subscript

    subscribed_name = generateChildExpressionCode(
        expression=subscribed, emit=emit, context=context
    )

    subscript_name = generateChildExpressionCode(
        expression=subscript, emit=emit, context=context
    )

    subscript_constant, integer_subscript = decideIntegerSubscript(subscript)

    with withObjectCodeTemporaryAssignment(
        to_name, "subscript_result", expression, emit, context
    ) as value_name:
        if integer_subscript:
            _getIntegerSubscriptLookupCode(
                to_name=value_name,
                subscribed_name=subscribed_name,
                subscript_name=subscript_name,
                subscript_value=subscript_constant,
                emit=emit,
                context=context,
            )
        else:
            _getSubscriptLookupCode(
                to_name=value_name,
                subscribed_name=subscribed_name,
                subscript_name=subscript_name,
                emit=emit,
                context=context,
            )


def generateMatchSubscriptCheckCode(to_name, expression, emit, context):
    subscribed = expression.subnode_expression
    subscript = expression.subnode_subscript

    subscribed_name = generateChildExpressionCode(
        expression=subscribed, emit=emit, context=context
    )

    subscript_name = generateChildExpressionCode(
        expression=subscript, emit=emit, context=context
    )

    res_name = context.getBoolResName()

    emit(
        "%s = MATCH_MAPPING_KEY(tstate, %s, %s);"
        % (res_name, subscribed_name, subscript_name)
    )

    getReleaseCodes((subscript_name, subscribed_name), emit, context)

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name,
        condition=res_name,
        emit=emit,
    )


def _getIntegerSubscriptLookupCode(
    to_name, subscribed_name, subscript_name, subscript_value, emit, context
):
    emit(
        "%s = LOOKUP_SUBSCRIPT_CONST(tstate, %s, %s, %s);"
        % (to_name, subscribed_name, subscript_name, subscript_value)
    )

    getErrorExitCode(
        check_name=to_name,
        release_names=(subscribed_name, subscript_name),
        emit=emit,
        context=context,
    )

    context.addCleanupTempName(to_name)


def _getSubscriptLookupCode(to_name, subscript_name, subscribed_name, emit, context):
    emit(
        "%s = LOOKUP_SUBSCRIPT(tstate, %s, %s);"
        % (to_name, subscribed_name, subscript_name)
    )

    getErrorExitCode(
        check_name=to_name,
        release_names=(subscribed_name, subscript_name),
        emit=emit,
        context=context,
    )

    context.addCleanupTempName(to_name)


def _getIntegerSubscriptAssignmentCode(
    subscribed_name, subscript_name, subscript_value, value_name, emit, context
):
    assert abs(subscript_value) < 2**31

    res_name = context.allocateTempName("ass_subscript_res", "int")

    emit(
        "%s = SET_SUBSCRIPT_CONST(tstate, %s, %s, %s, %s);"
        % (res_name, subscribed_name, subscript_name, subscript_value, value_name)
    )

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_names=(subscribed_name, value_name),
        emit=emit,
        context=context,
    )


def _getSubscriptAssignmentCode(target_name, subscript_name, value_name, emit, context):
    res_name = context.getBoolResName()

    emit(
        "%s = SET_SUBSCRIPT(tstate, %s, %s, %s);"
        % (res_name, target_name, subscript_name, value_name)
    )

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_names=(target_name, subscript_name, value_name),
        emit=emit,
        context=context,
    )


def _getSubscriptDelCode(target_name, subscript_name, emit, context):
    res_name = context.getBoolResName()

    emit("%s = DEL_SUBSCRIPT(%s, %s);" % (res_name, target_name, subscript_name))

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_names=(target_name, subscript_name),
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code generation for match statement helpers."""

from .CodeHelpers import (
    generateChildExpressionsCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitCode


def generateMatchArgsCode(to_name, expression, emit, context):
    (matched_value_name, match_type_name) = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    # TODO: Prefer "PyObject **" of course once we have that.
    keywords = expression.getKeywordArgs()

    if keywords:
        keywords_name = context.getConstantCode(constant=keywords)
        keywords_name = "&PyTuple_GET_ITEM(%s, 0)" % keywords_name
    else:
        keywords_name = "NULL"

    with withObjectCodeTemporaryAssignment(
        to_name, "match_args_value", expression, emit, context
    ) as value_name:
        emit(
            "%s = MATCH_CLASS_ARGS(tstate, %s, %s, %d, %s, %d);"
            % (
                value_name,
                matched_value_name,
                match_type_name,
                expression.getPositionalArgsCount(),
                keywords_name,
                len(keywords),
            )
        )

        getErrorExitCode(
            check_name=value_name,
            release_names=(matched_value_name, match_type_name),
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(value_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
